
//
// 8-button GKOS input method for mobile phone chorded keyboard
// Losely based on the outdated library of Seppo Tiainen, 28-Mar-2010
// Changed implementation to interface with Leonardo Keyboard library;
// Reduced memory consumption from 52 to 35% while adding a national character set.
// Key assignments changed as per the letter frequency.
// v 2.2 04-May-2018
// This work is public domain
//

//#include <LowPower.h>
#include <Keyboard.h>

// Uncomment for serial debugging
//#define _DEBUG
//#define _DISABLE_KEYBOARD

#define _CH0 (char)0

// Special non-printable characters are in place of F-buttons and PrintScreen 0xC1-0xCE

#define _RTN (char)0xB0
#define _ESC (char)0xB1
#define _BSP (char)0xB2
#define _TAB (char)0xB3

#define _SH_ (char)0xC1
#define _CH_ (char)0xC2
#define _AND (char)0xC3
#define _OR_ (char)0xC4
#define _YOU (char)0xC5
#define _TH_ (char)0xC6
#define _THE (char)0xC7
#define _COP (char)0xC8
#define _PAS (char)0xC9
#define _RUS (char)0xCA
#define _CUT (char)0xCB
#define _UDO (char)0xCD
#define _ANB (char)0xCE
#define _ABB (char)0xCF

#define _INS (char)0xD1
#define _HOM (char)0xD2
#define _PUP (char)0xD3
#define _DEL (char)0xD4
#define _END (char)0xD5
#define _PDN (char)0xD6
#define _RT_ (char)0xD7
#define _LT_ (char)0xD8
#define _DN_ (char)0xD9
#define _UP_ (char)0xDA

// Fake Control arrows
#define _CRT (char)0xE7
#define _CLT (char)0xE8
#define _CDN (char)0xE9
#define _CUP (char)0xEA

// Fake Shift and shift-control arrows
#define _SRT (char)0xF7
#define _SLT (char)0xF8
#define _SDN (char)0xF9
#define _SUP (char)0xFA
#define _CSR (char)0xFB
#define _CSL (char)0xFC

#define _CTRL (char)0x80
#define _SHIFT (char)0x81
#define _ALT (char)0x82
#define _GUI (char)0x83
#define _CTRL_R (char)0x84
#define _SHIFT_R (char)0x85
#define _ALT_R (char)0x86
#define _RIGHT_R (char)0x87

boolean keyboardActive = false;                    // Safety feature, in case programming goes bananas

//
// Pin assignment
//
//const int GKOSPins[] = {4, 3, 2, 6, 5, 7, 8, 9};     // reading pins for 6 GKOS buttons and 2 mode buttons
const int GKOSPins[] = {4, 5, 16, 10, 14, 15, 6, 2};   // reading pins for 6 GKOS buttons and 2 mode buttons
const int LEDPins[] = {7, 3};                          // LED pins for indicators

//
// Character definition
//
// English Lower Case
const static char GKOS_Char[]={
  _CH0,   'e',   'a',   't',    'l',   'p',   's',  _BSP,
   'i',  _UP_,   'j',   'v',    'u',  _SH_,   'b',  _RT_,
   'o',   'm',  _RTN,   'w',    ',',  _CH_,   'c',  _SRT,
   'n',   'g',   'q',  _PUP,   _AND,  _OR_,   '?',  _CRT,
   'r',   'd',   '.',   'z',   _DN_,   '-',   'k',  _END,
   'y',  _YOU,  _TH_,  _SUP,   _THE,  _DEL,  _SDN,  _COP,
   'h',   'f',   'x',   '!',    '"',   '#',  _PDN,  _CSR,
   ' ',  _LT_, _SLT,   _CLT,   _HOM,  _PAS,  _CSL,  _RUS};

// English Upper Case
const static char GKOS_Caps[]={
  _CH0,   'E',   'A',   'T',    'L',   'P',   'S',  _BSP,
   'I',  _UP_,   'J',   'V',    'U',  _SH_,   'B',  _RT_,
   'O',   'M',  _RTN,   'W',    ';',  _CH_,   'C',  _SRT,
   'N',   'G',   'Q',  _PUP,   _AND,  _OR_,   '?',  _CRT,
   'R',   'D',   ':',   'Z',   _DN_,   '_',   'K',  _END,
   'Y',  _YOU,  _TH_,  _SUP,   _THE,  _DEL,  _SDN,  _COP,
   'H',   'F',   'X',   '!',    '"',   '#',  _PDN,  _CSR,
   ' ',  _LT_, _SLT,   _CLT,   _HOM,  _PAS,  _CSL,  _RUS};

// Russian Lower Case
const static char GKOS_Char_Rus[]={
  _CH0,   't',   'f',   'n',    'k',   'g',   'c',  _BSP,
   'b',  _UP_,   'q',   'd',    'e',   'i',   ',',  _RT_,
   'j',   'v',  _RTN,   ';',    '?',   'o',   'w',  _SRT,
   'y',   'u',   'z',  _PUP,    'm',   ']',   '&',  _CRT,
   'h',   'l',   '/',   'p',   _DN_,   '-',   'r',  _END,
   's',   '.',  '\'',  _SUP,    '`',  _DEL,  _SDN,  _COP,
   'x',   'a',   '[',   '!',    '@',   '#',  _PDN,  _CSR,
   ' ',  _LT_, _SLT,   _CLT,   _HOM,  _PAS,  _CSL,  _RUS};

// Russian Upper Case
const static char GKOS_Caps_Rus[]={
  _CH0,   'T',   'F',   'N',    'K',   'G',   'C',  _BSP,
   'B',  _UP_,   'Q',   'D',    'E',   'I',   '<',  _RT_,
   'J',   'V',  _RTN,   ':',    '$',   'O',   'W',  _SRT,
   'Y',   'U',   'Z',  _PUP,    'M',   '}',   '&',  _CRT,
   'H',   'L',   '^',   'P',   _DN_,   '_',   'R',  _END,
   'S',   '>',  '\"',  _SUP,    '~',  _DEL,  _SDN,  _COP,
   'X',   'A',   '{',   '!',    '@',   '#',  _PDN,  _CSR,
   ' ',  _LT_, _SLT,   _CLT,   _HOM,  _PAS,  _CSL,  _RUS};
      
const static char GKOS_SYMB[]={
  _CH0,   '1',   '2',   '+',    '3',   '(',   '=',  _BSP,
   '4',  _UP_,   '|',   '-',    '%',   '[',   '<',  _RT_,
   '5',   '&',  _ESC,   '/',    ',',   '{',   '>',  _SRT,
   '0',   '7',   '8',  _PUP,    '9',  _ABB,   '$',  _CRT,
   '6',   '^',   '.',   '*',   _DN_,   '-',   '%',  _END,
   ')',   ']',   '}',  _SUP,    '`',  _INS,  _SDN,  _CUT,
   'e',   '>',   '~',   '@',    '"',   '#',  _PDN,  _CSR,
  _TAB,  _LT_, _SLT,   _CLT,   _HOM,  _UDO,  _CSL,  _RUS};
   
const static char GKOS_Nums[]={
  _CH0,   '1',   '2',   '+',    '3',   '(',   '=',  _BSP,
   '4',  _UP_,   '|',   '-',    '%',   '[',   '<',  _RT_,
   '5',   '&',  _RTN,   '/',    ',',   '{',   '>',  _SRT,
   '0',   '7',   '8',  _PUP,    '9',  _ANB,   '$',  _CRT,
   '6',   '^',   '.',   '*',   _DN_,   '-',   '%',  _END,
   ')',   ']',   '}',  _SUP,    '`',  _DEL,  _SDN,  _COP,
   'e',   '>',   '~',   '@',    '"',   '#',  _PDN,  _CSR,
   ' ',  _LT_, _SLT,   _CLT,   _HOM,  _UDO,  _CSL,  _RUS};

const static char _Pars0_Substitute[] = {'(', ')', _LT_, _CH0}; 
const static char _Pars1_Substitute[] = {'\"', '\"', _LT_, _CH0}; 
const static char _Pars2_Substitute[] = {'@', '@', _LT_, _CH0}; 
const static char _Pars3_Substitute[] = {'[', ']', _LT_, _CH0}; 
const static char _Pars4_Substitute[] = {'{', '}', _LT_, _CH0}; 
const static char _Pars5_Substitute[] = {'<', '>', _LT_, _CH0}; 
const static char _Pars6_Substitute[] = {'<', '/', '>', _LT_, _LT_, _CH0}; 
const static char _Pars7_Substitute[] = {' ', '|', '|', ' ', _CH0}; 
const static char _Pars8_Substitute[] = {' ', '&', '&', ' ', _CH0}; 

//
// Buttons data
//
long  GKOS_time = 0;
long  GKOS_previousTime = 0;
long  GKOS_autoRepeat = 0;     // Repeat the character if chord pressed long enough (0 or 1)
int   GKOS_presentChord = 0;   // The chord of the keypad at present
int   GKOS_previousChord = 0;  // To compare chord change
int   GKOS_autoCounter = 0;    // Typamatic delay counter (n x 10 ms)
int   GKOS_firstChord = 0;     // First chord of a Chordon, if any
int   GKOS_firstCounter = 0;   // Count to detect a separate first chord in a Chordon
int   GKOS_gChord = 0;         // GKOS final chord value (0...63) for the character
bool  GKOS_gNew = false;       // a new character is expected soon because new keys were pressed
bool  GKOS_prevLeft = false;   // previous value of the left (front) mode button
bool  GKOS_prevRight = false;  // previous value of the left (front) mode button

bool GKOS_Russian = false;     // indicates if the keyboard in Russian mode
bool GKOS_prevRussian = false;
int  GKOS_mode = 0;            // mode: 0 - normal, 1 - shift, 2 - caps lock, 3 - sym, 4 - num lock
char GKOS_output = ' ';

//
// Control keys; under Android Win key is useless anyway
//
bool CTRL_Requested = false;
bool ALT_Requested = false;

//
// Setup code here, to run once:
//
void setup()
{
  #ifdef _DEBUG
  Serial.begin(9600);
  #endif

  for ( int i = 0; i < 2; i++)
  {
    pinMode(LEDPins[i], OUTPUT);
    digitalWrite(LEDPins[i], HIGH);
  }
  for ( int i = 0; i < 8; i++) pinMode(GKOSPins[i], INPUT_PULLUP);
  keyboardActive = !(ReadPinBool(6) || ReadPinBool(7));
  if ( !keyboardActive) return;
  
  #ifndef _DISABLE_KEYBOARD
  Keyboard.begin();
  #endif
  
  delay( 500);
  SetLEDs();
}

//
// Runs continuously
//
void loop()
{
  bool output_ready;
  if ( keyboardActive)
  {
    SetModeButtons();
    if (GKOS_Read_Buttons())
    {
      output_ready = GKOS_Decoder();
      #ifdef _DEBUG
      if (output_ready) Serial.println( "Button code retrieved: " + String( GKOS_output, HEX));
      #endif
      if (output_ready) SendButtons();
    }
  }
  #ifdef _DEBUG
  delay(500);
  #else
  // Eliminate unwanted Tx, Rx lights
  RXLED1;
  TXLED1;
  delay(15);
  //LowPower.powerDown(SLEEP_15MS, ADC_OFF, BOD_OFF);
  #endif
  return;
}

//
// Reads mode buttons, sets the mode
//
void SetModeButtons()
{
  bool modebutton = ReadPinBool( 7);
  if( modebutton && !GKOS_prevRight)
  {
    GKOS_prevRight = true;
    switch (GKOS_mode)
    {
      case 3:
        GKOS_mode = 4;
        #ifdef _DEBUG
        Serial.println( "Mode set to NUM");
        #endif
        break;
      case 4:
        GKOS_mode = 0;
        ForceRusLat( GKOS_prevRussian);
        #ifdef _DEBUG
        Serial.println( "Mode set to NORMAL");
        #endif
        break;
      default:
        GKOS_mode = 3;
        GKOS_prevRussian = GKOS_Russian; 
        ForceRusLat( false);
        #ifdef _DEBUG
        Serial.println( "Mode set to SYMBOL");
        #endif
        break;
    }
    SetLEDs();
  }  
  if( !modebutton && GKOS_prevRight) GKOS_prevRight = false;
  modebutton = ReadPinBool( 6);
  if( modebutton && !GKOS_prevLeft)
  {
    GKOS_prevLeft = true;
    switch (GKOS_mode)
    {
      case 0:
        GKOS_mode = 1;
        #ifdef _DEBUG
        Serial.println( "Mode set to SHIFT");
        #endif
        break;
      case 1:
        GKOS_mode = 2;
        #ifdef _DEBUG
        Serial.println( "Mode set to CAPS");
        #endif
        break;
      case 3:
        GKOS_mode = 4;
        #ifdef _DEBUG
        Serial.println( "Mode set to NUM");
        #endif
        break;
      case 4:
        GKOS_mode = 0;
        ForceRusLat( GKOS_prevRussian);
        #ifdef _DEBUG
        Serial.println( "Mode set to NORMAL");
        #endif
        break;
      default:
        GKOS_mode = 0;
        #ifdef _DEBUG
        Serial.println( "Mode set to NORMAL");
        #endif
        break;
    }
    SetLEDs();
  }  
  if( !modebutton && GKOS_prevLeft) GKOS_prevLeft = false;
}

//
// Reads the rear 6 buttons, processes them into the press codes
//
bool GKOS_Read_Buttons()
{
  int button_weight = 1;
  int pin = 0;
  int i, j;

  // Return empty immediately if 10 ms not elapsed
  GKOS_time = millis();
  if ((GKOS_time - GKOS_previousTime) < (10 + 64 * GKOS_autoRepeat)) return false;

  // Read buttons and form a code 
  GKOS_presentChord = 0;
  for( int i=0; i<6; i++)
  {
    pin = ReadPin( i);
    GKOS_presentChord += pin * button_weight;
    button_weight += button_weight; 
  }

  // any buttons pressed
  if (GKOS_presentChord != 0)
  {
    // Chord pressed long enough for autorepeat
    GKOS_autoCounter++;
    if (GKOS_autoCounter >= 100)
    {
      GKOS_autoCounter = 99;
      GKOS_autoRepeat = 1;
    }

    // Chord pressed long enough to be first chord in Chordon
    GKOS_firstCounter++;
    if (GKOS_firstCounter > 25)
    {
      GKOS_firstCounter = 25;
      if (GKOS_firstChord == 0) GKOS_firstChord = GKOS_presentChord;
    }
  }

  // New key(s) pressed, expect a new character soon
  if (GKOS_presentChord > GKOS_previousChord)
  {
    GKOS_gNew = true;
    GKOS_autoCounter = 0;
    GKOS_autoRepeat = 0;
    GKOS_firstCounter = 0;
  }

  // key(s) relased
  if (GKOS_presentChord < GKOS_previousChord)
  {
    GKOS_autoCounter = 0;
    GKOS_autoRepeat = 0;
    GKOS_firstCounter = 0;
    GKOS_gChord = GKOS_previousChord;  // the chord before the first release of key(s) is what we are after
    GKOS_previousChord = GKOS_presentChord;  
    GKOS_previousTime = GKOS_time;
    if (!GKOS_gNew) return false;
    GKOS_gNew = false; // next key releases will no more represent chords for characters
    return true; // a real chord entered!
  }

  // no buttons pressed
  if (GKOS_presentChord == 0)
  {
    GKOS_firstChord = 0; 
    GKOS_firstCounter = 0;
  }
  GKOS_previousChord = GKOS_presentChord; // to be able to compare next time
  GKOS_previousTime = GKOS_time;

  if (!GKOS_autoRepeat) return false;
  GKOS_gChord = GKOS_previousChord;
  GKOS_firstCounter = 0;
  GKOS_firstChord = 0;
  return true;
}

//
// Decodes the buttons based on the current mode
//
bool GKOS_Decoder()
{
  // If it is a Chordon, send it here
  // Seppo Tiainen did not finish this
  /*     if (_firstChord != 0){_firstChord = 0; _firstCounter = 0; // TEMP!!!!
        return "-";} // TODO: combine two characters and return directly */
  GKOS_firstChord = 0;
  GKOS_firstCounter = 0;  // in any case, delete the first char of Chordon
  switch (GKOS_mode)
  {
    case 0: // Normal
      if( GKOS_Russian) GKOS_output = GKOS_Char_Rus[GKOS_gChord];
      else GKOS_output = GKOS_Char[GKOS_gChord];
      break;
    case 1: // Shift
    case 2: // Caps 
      if( GKOS_Russian) GKOS_output = GKOS_Caps_Rus[GKOS_gChord];
      else GKOS_output = GKOS_Caps[GKOS_gChord];
      break;
    case 3: // Symbols 
      GKOS_output = GKOS_SYMB[GKOS_gChord];
    default: // Nums
      GKOS_output = GKOS_Nums[GKOS_gChord];
      break;
  } // switch
  return true;
}

void SendButtons()
{
  if( GKOS_output == _CH0) return;
  if( SubstituteCtrl(_CTRL_R)) return;
  if( SubstituteAlt(_ALT_R)) return;
  if( SubstituteRusLat()) return;
  if( SubstitutePunctuation()) return;

  // Change metacodes to words
  if( SubstituteMetacode( _SH_, "sh", "Sh", "SH")) return;
  if( SubstituteMetacode( _SH_, "sh", "Sh", "SH")) return;
  if( SubstituteMetacode( _CH_, "ch", "Ch", "CH")) return;
  if( SubstituteMetacode( _AND, "and", "And", "AND")) return;
  if( SubstituteMetacode( _OR_, "or", "Or", "OR")) return;
  if( SubstituteMetacode( _YOU, "you", "You", "YOU")) return;
  if( SubstituteMetacode( _TH_, "th", "Th", "TH")) return;
  if( SubstituteMetacode( _THE, "the", "The", "THE")) return;
  
  // Substitute for double brackets
  if( SubstituteMetacodeArray( '(', _Pars0_Substitute, false, false, false))  return;
  if( SubstituteMetacodeArrayEnglish( '\"', _Pars1_Substitute, false, false, false))  return;
  if( SubstituteMetacodeArrayRussian( '@', _Pars2_Substitute, false, false, false))  return;
  if( SubstituteMetacodeArrayEnglish( '[', _Pars3_Substitute, false, false, false))  return;
  if( SubstituteMetacodeArrayEnglish( '{', _Pars4_Substitute, false, false, false))  return;
  if( SubstituteMetacodeArraySymbol( _ANB, _Pars5_Substitute, false, false, false))  return;
  if( SubstituteMetacodeArraySymbol( _ABB, _Pars6_Substitute, false, false, false))  return;
  if( SubstituteMetacodeArrayNumber( _ANB, _Pars5_Substitute, false, false, false))  return;
  if( SubstituteMetacodeArrayNumber( _ABB, _Pars6_Substitute, false, false, false))  return;
  if( SubstituteMetacodeArraySymbol( '|', _Pars7_Substitute, false, false, false))  return;
  if( SubstituteMetacodeArraySymbol( '&', _Pars8_Substitute, false, false, false))  return;

  // substitute for control and shift characters
  if( SubstituteMetacodeCharacter( _COP, 'c', false, true, false)) return;
  if( SubstituteMetacodeCharacter( _PAS, 'v', false, true, false)) return;
  if( SubstituteMetacodeCharacter( _CUT, 'x', false, true, false)) return;
  if( SubstituteMetacodeCharacter( _UDO, 'z', false, true, false)) return;
  if( SubstituteMetacodeCharacter( _CRT, _RT_, false, true, false)) return;
  if( SubstituteMetacodeCharacter( _CLT, _LT_, false, true, false)) return;
  if( SubstituteMetacodeCharacter( _CDN, _DN_, false, true, false)) return;
  if( SubstituteMetacodeCharacter( _CUP, _UP_, false, true, false)) return;
  if( SubstituteMetacodeCharacter( _SRT, _RT_, false, false, true)) return;
  if( SubstituteMetacodeCharacter( _SLT, _LT_, false, false, true)) return;
  if( SubstituteMetacodeCharacter( _SDN, _DN_, false, false, true)) return;
  if( SubstituteMetacodeCharacter( _SUP, _UP_, false, false, true)) return;
  if( SubstituteMetacodeCharacter( _CSR, _RT_, false, true, true)) return;
  if( SubstituteMetacodeCharacter( _CSL, _LT_, false, true, true)) return;

  if( CTRL_Requested) Keyboard_press( _CTRL_R);
  if( ALT_Requested) Keyboard_press( _ALT_R);
  Keyboard_write( GKOS_output);
  if( ALT_Requested) Keyboard_release( _ALT_R);
  if( CTRL_Requested) Keyboard_release( _CTRL_R);
  if( (GKOS_mode == 1 || GKOS_mode == 3) && GKOS_output != _RTN)
  {
    if( GKOS_mode == 3) ForceRusLat( GKOS_prevRussian);
    GKOS_mode = 0;
    SetLEDs();
  }
  CTRL_Requested = false;
  ALT_Requested = false;
}

//
// Substitutes a mata-code with a key code; returns true if further processing not required
//
bool SubstituteMetacode( char code, String meta, String meta_shift, String meta_cap)
{  
  if( code != GKOS_output) return false;
  switch( GKOS_mode)
  {
    case 0:
      Keyboard_print( meta);
      return true;
    case 1:
      Keyboard_print( meta_shift);
      GKOS_mode = 0;
      SetLEDs();
      return true;
    case 2:
      Keyboard_print( meta_cap);
      return true;
    default:
      break;
  }
  return false;
}

//
// Substitutes a mata-code with a key code; returns true if further processing not required
//
bool SubstituteMetacodeCharacter( char code, char meta, bool Alt, bool Ctrl, bool Shift)
{  
  if( code != GKOS_output) return false;
  if( Alt) Keyboard_press( _ALT);
  if( Ctrl) Keyboard_press( _CTRL);
  if( Shift) Keyboard_press( _SHIFT);
  Keyboard_write( meta);
  if( Shift) Keyboard_release( _SHIFT);
  if( Ctrl) Keyboard_release( _CTRL);
  if( Alt) Keyboard_release( _ALT);
  return true;
}

//
// Substitutes a mata-code with a sequence from a zero-terminated array; returns true if further processing not required
//
bool SubstituteMetacodeArray( char code, const char meta[], bool Alt, bool Ctrl, bool Shift)
{  
  if( code != GKOS_output) return false;
  if( Alt) Keyboard_press( _ALT);
  if( Ctrl) Keyboard_press( _CTRL);
  if( Shift) Keyboard_press( _SHIFT);
  for( int i=0; i<100; i++)
  {
    if( meta[i] == _CH0) break;
    Keyboard_write( meta[i]);
  } 
  if( Shift) Keyboard_release( _SHIFT);
  if( Ctrl) Keyboard_release( _CTRL);
  if( Alt) Keyboard_release( _ALT);
  return true;
}

//
// Substitutes a mata-code with a sequence from a zero-terminated array; returns true if further processing not required
//
bool SubstituteMetacodeArrayEnglish( char code, const char meta[], bool Alt, bool Ctrl, bool Shift)
{
  if( GKOS_Russian) return false;
  return SubstituteMetacodeArray( code, meta, Alt, Ctrl, Shift);
}

//
// Substitutes a mata-code with a sequence from a zero-terminated array; returns true if further processing not required
//
bool SubstituteMetacodeArrayRussian( char code, const char meta[], bool Alt, bool Ctrl, bool Shift)
{
  if( !GKOS_Russian) return false;
  return SubstituteMetacodeArray( code, meta, Alt, Ctrl, Shift);
}

//
// Substitutes a mata-code with a sequence from a zero-terminated array; returns true if further processing not required
//
bool SubstituteMetacodeArraySymbol( char code, const char meta[], bool Alt, bool Ctrl, bool Shift)
{
  if( GKOS_mode !=3 ) return false;
  return SubstituteMetacodeArray( code, meta, Alt, Ctrl, Shift);
}

//
// Substitutes a mata-code with a sequence from a zero-terminated array; returns true if further processing not required
//
bool SubstituteMetacodeArrayNumber( char code, const char meta[], bool Alt, bool Ctrl, bool Shift)
{
  if( GKOS_mode != 4) return false;
  return SubstituteMetacodeArray( code, meta, Alt, Ctrl, Shift);
}

//
// Prints one substitute string
//
bool PrintSubstitute( char code, String s, int newMode)
{
  if( code != GKOS_output) return false;
  Keyboard_print( s);
  if( newMode == GKOS_mode) return true;
  GKOS_mode = newMode;
  SetLEDs();
  return true;
}

//
// Substitutes punctuation in modes 0, 1, 2
//
bool SubstituteEnglishPunctuation()
{
  switch(GKOS_mode)
  {
    case 0:
    case 1:
      if( PrintSubstitute( ',', ", ", 0)) return true;
      if( PrintSubstitute( '.', ". ", 1)) return true;
      if( PrintSubstitute( '?', "? ", 1)) return true;
      if( PrintSubstitute( '!', "! ", 1)) return true;
      if( PrintSubstitute( ';', "; ", 0)) return true;
      if( PrintSubstitute( ':', ": ", 0)) return true;
      return false;
    case 2:
      if( PrintSubstitute( ',', ", ", 2)) return true;
      if( PrintSubstitute( '.', ". ", 2)) return true;
      if( PrintSubstitute( '?', "? ", 2)) return true;
      if( PrintSubstitute( '!', "! ", 2)) return true;
      if( PrintSubstitute( ';', "; ", 2)) return true;
      if( PrintSubstitute( ':', ": ", 2)) return true;
      return false;
    default:
      break;
  }
  return false;
}

//
// Substitutes punctuation in modes 0, 1, 2
//
bool SubstituteRussianPunctuation()
{
  switch(GKOS_mode)
  {
    case 0:
    case 1:
      if( PrintSubstitute( '/', "/ ", 1)) return true;
      if( PrintSubstitute( '?', "? ", 0)) return true;
      if( PrintSubstitute( '$', "$ ", 0)) return true;
      if( PrintSubstitute( '^', "^ ", 1)) return true;
      if( PrintSubstitute( '&', "& ", 1)) return true;
      if( PrintSubstitute( '!', "! ", 1)) return true;
      return false;
    case 2:
      if( PrintSubstitute( '/', "/ ", 2)) return true;
      if( PrintSubstitute( '?', "? ", 2)) return true;
      if( PrintSubstitute( '$', "$ ", 2)) return true;
      if( PrintSubstitute( '^', "^ ", 2)) return true;
      if( PrintSubstitute( '&', "& ", 2)) return true;
      if( PrintSubstitute( '!', "! ", 2)) return true;
      return false;
    default:
      break;
  }
  return false;
}

//
// Substitutes punctuation in modes 0, 1, 2
//
bool SubstitutePunctuation()
{
  if( GKOS_Russian) return SubstituteRussianPunctuation();
  else return SubstituteEnglishPunctuation();
}

//
// Sets Control key
//
bool SubstituteCtrl( char code)
{
  if( GKOS_output != code) return false;
  CTRL_Requested = true;
  return true;
}

//
// Sets Alt key
//
bool SubstituteAlt( char code)
{
  if( GKOS_output != code) return false;
  ALT_Requested = true;
  return true;
}

//
// Changes between Rus and Lat by pressing and releasing Shift-Control
//
bool SubstituteRusLat()
{
  if( GKOS_output != _RUS) return false;
  Keyboard_press( _CTRL_R);
  Keyboard_write( ' ');
  Keyboard_release( _CTRL_R);
  if( GKOS_mode < 3) GKOS_Russian = !GKOS_Russian;
  return true;
}

//
// Explicit change between Rus and Lat
//
void ForceRusLat( bool rus)
{
  if( rus == GKOS_Russian) return;
  Keyboard_press( _CTRL_R);
  Keyboard_write( ' ');
  Keyboard_release( _CTRL_R);
  GKOS_Russian = rus;
}

//
// Low-level functions to read pins and set LEDs
//
//
// Reads a pin, returns the reading
//
int ReadPin( int i)
{
  if ( digitalRead( GKOSPins[i]) == LOW) return 1;
  return 0;
}

//
// Reads a pin, returns the reading
//
bool ReadPinBool( int i)
{
  return digitalRead( GKOSPins[i]) == LOW;
}

//
// Sets the LEDs based on the mode
//
void SetLEDs()
{
  switch( GKOS_mode)
  {
    case 0:
      WriteLEDs( false, false);
      break;
    case 1:
    case 2:
      WriteLEDs( true, false);
      break;
    case 3:
      WriteLEDs( false, true);
      break;
    case 4:
      WriteLEDs( true, true);
      break;
  }
}

//
// Writes to LEDs
//
void WriteLEDs( bool v1, bool v2)
{
  digitalWrite(LEDPins[0], v1);
  digitalWrite(LEDPins[1], v2);
}

//
// Low-level functions to handle keyboard outputs
//
void Keyboard_press(char key)
{
  #ifndef _DISABLE_KEYBOARD
  Keyboard.press( key);
  #endif
}

void Keyboard_write(char key)
{
  #ifndef _DISABLE_KEYBOARD
  Keyboard.write( key);
  #endif
}

void Keyboard_print(String s)
{
  #ifndef _DISABLE_KEYBOARD
  Keyboard.print( s);
  #endif
}

void Keyboard_release(char key)
{
  #ifndef _DISABLE_KEYBOARD
  Keyboard.release( key);
  #endif
}

